export const generateQuery = (function () {
    function generateQuery(definition, options) {
        const opts = {
            indent: (options || {}).indent || 0,
            indentChar: (options || {}).indentChar
        };
        const doc = [];
        const parse = (type) => {
            const op = definition[type];
            if (op) {
                let code = createIndent(opts.indent, opts.indentChar) + type;
                code = append(code, op.$name);
                const { $variables, $directives, $fields } = op;
                if ($variables) {
                    const vars = [];
                    const varIndent = createIndent(opts.indent + 1, opts.indentChar);
                    Object.keys($variables).forEach((key) => {
                        const val = $variables[key];
                        if (val) {
                            vars.push(varIndent + key + ': ' + val);
                        }
                    });
                    if (vars.length) {
                        code +=
                            ' (\n' +
                                vars.join('\n') +
                                '\n' +
                                createIndent(opts.indent, opts.indentChar) +
                                ')';
                    }
                }
                if ($directives) {
                    code = append(code, encodeDirectives($directives, opts.indent, opts.indentChar));
                }
                const fields = [];
                if ($fields) {
                    $fields.forEach((field) => fields.push(field));
                }
                fields.push(op);
                const body = encodeFields(fields, opts);
                if (body) {
                    doc.push(code + ' ' + body);
                }
            }
        };
        parse('query');
        parse('mutation');
        parse('subscription');
        each(definition.fragments, (fragments) => {
            const code = encodeFragments(fragments, opts);
            if (code) {
                doc.push(code);
            }
        });
        return doc.join('\n\n');
    }
    function encodeFragments(fragments, options) {
        const code = [];
        const indent = createIndent(options.indent, options.indentChar);
        Object.keys(fragments).forEach((key) => {
            const fragment = fragments[key];
            if (fragment) {
                const body = encodeFields(fragment, options);
                if (body) {
                    code.push(append(indent + 'fragment ' + key + ' on ' + fragment.$onType, encodeDirectives(fragment.$directives, options.indent, options.indentChar)) +
                        ' ' +
                        body);
                }
            }
        });
        return code.join('\n\n');
    }
    function encodeFields(fields, options) {
        const { indent, indentChar } = options;
        const code = [];
        const propIndent = createIndent(indent + 1, indentChar);
        const encodeSubField = (key, subField) => {
            if (subField.$content) {
                code.push(propIndent + subField.$content);
            }
            else {
                let prop = '';
                if (subField.$alias) {
                    prop = subField.$alias + ': ';
                }
                prop += key;
                if (subField.$args) {
                    prop = append(prop, encodeArgs(subField.$args, {
                        indent: indent + 1,
                        indentChar
                    }));
                }
                if (subField.$directives) {
                    prop = append(prop, encodeDirectives(subField.$directives, indent + 1, indentChar));
                }
                if (subField.$body) {
                    prop += ' ' + subField.$body;
                }
                else {
                    prop = append(prop, encodeFields(subField, {
                        indent: indent + 1,
                        indentChar
                    }));
                }
                code.push(propIndent + prop);
            }
        };
        each(fields, (field) => {
            Object.keys(field).forEach((key) => {
                const val = field[key];
                if (val) {
                    if (key[0] === '$') {
                        if (key === '$on') {
                            each(Object.keys(val), (typeName) => {
                                each(val[typeName], (typeBody) => {
                                    const inlineFragment = propIndent +
                                        (typeName === '$' ? '...' : '... on ' + typeName);
                                    const fragmentBody = encodeFields(typeBody, {
                                        indent: indent + 1,
                                        indentChar
                                    });
                                    if (fragmentBody) {
                                        code.push(append(inlineFragment, encodeDirectives(typeBody.$directives, indent + 1, indentChar)) +
                                            ' ' +
                                            fragmentBody);
                                    }
                                });
                            });
                        }
                        else if (key === '$spread') {
                            each(val, (item) => {
                                let name;
                                let directives;
                                if (typeof item === 'string') {
                                    name = item;
                                }
                                else {
                                    name = item.name;
                                    directives = item.directives;
                                }
                                code.push(append(propIndent + '...' + name, encodeDirectives(directives, indent + 1, indentChar)));
                            });
                        }
                    }
                    else if (typeof val === 'boolean' || typeof val === 'number') {
                        code.push(propIndent + key);
                    }
                    else if (typeof val === 'string') {
                        code.push(propIndent + val + ': ' + key);
                    }
                    else if (Array.isArray(val)) {
                        val.forEach((subField) => encodeSubField(key, subField));
                    }
                    else {
                        encodeSubField(key, val);
                    }
                }
            });
        });
        return code.length
            ? '{\n' + code.join('\n') + '\n' + createIndent(indent, indentChar) + '}'
            : '';
    }
    function encodeArgs(args, options) {
        const { object, indent, indentChar } = options;
        const propIndent = createIndent(indent + 1, indentChar);
        const code = [];
        let keep = false;
        Object.keys(args).forEach((key) => {
            if (key[0] === '$') {
                if (key === '$keep' && object && args[key]) {
                    keep = true;
                }
            }
            else {
                const val = getArgValue(args[key], indent, indentChar);
                if (val) {
                    code.push(propIndent + key + ': ' + val);
                }
            }
        });
        if (code.length) {
            const body = code.join('\n') + '\n' + createIndent(indent, indentChar);
            return object ? '{\n' + body + '}' : '(\n' + body + ')';
        }
        return keep ? '{}' : '';
    }
    function getArgValue(value, indent, indentChar) {
        if (value !== undefined) {
            const type = typeof value;
            if (type === 'number' || type === 'boolean' || value === null) {
                return '' + value;
            }
            else if (type === 'string') {
                return JSON.stringify(value);
            }
            else if (Array.isArray(value)) {
                const list = [];
                value.forEach((elem) => {
                    const item = getArgValue(elem, indent, indentChar);
                    if (item) {
                        list.push(item);
                    }
                });
                return '[' + list.join(', ') + ']';
            }
            else {
                const keys = ['$enum', '$var', '$raw'];
                for (let i = 0; i < keys.length; i += 1) {
                    if (value[keys[i]] !== undefined) {
                        return '' + value[keys[i]];
                    }
                }
                return encodeArgs(value, {
                    indent: indent + 1,
                    object: true,
                    indentChar
                });
            }
        }
        return '';
    }
    function encodeDirectives(directive, indent, indentChar) {
        if (typeof directive === 'string') {
            return directive;
        }
        else if (Array.isArray(directive)) {
            return directive
                .map((el) => encodeDirectives(el, indent, indentChar))
                .join(' ');
        }
        else if (directive) {
            let code = directive.name || '';
            if (code && directive.args) {
                code = append(code, encodeArgs(directive.args, {
                    indent,
                    indentChar
                }));
            }
            return code;
        }
        return '';
    }
    function createIndent(level, char) {
        return Array(level + 1).join(char || '  ');
    }
    function each(arr, callback) {
        if (Array.isArray(arr)) {
            arr.forEach(callback);
        }
        else if (arr) {
            callback(arr);
        }
    }
    function append(to, value) {
        return value ? to + ' ' + value : to;
    }
    return generateQuery;
})();
